import math
from scipy.stats import beta
from estimators.bandits import base

class Interval(base.Interval):

    def __init__(self):
        ################################# Aggregates quantities #########################################
        #
        # 'n':   IPS of numerator
        # 'N':   total number of samples in bin from log (IPS = n/N)
        # 'c':   max abs. value of numerator's items (needed for Clopper-Pearson confidence intervals)
        #
        #################################################################################################

        self.data = {'n':0.,'N':0,'c':0.}

    def add_example(self, p_log, r, p_pred, count=1):
        self.data['N'] += count
        if p_pred > 0:
            p_over_p = p_pred/p_log
            if r != 0:
                self.data['n'] += r*p_over_p*count
                self.data['c'] = max(self.data['c'], r*p_over_p)

    def get(self, alpha=0.05):
        bounds = []
        num = self.data['n']
        den = self.data['N']
        maxWeightedCost = self.data['c']

        if maxWeightedCost > 0.0:
            successes = num / maxWeightedCost
            n = den / maxWeightedCost
            bounds.append(beta.ppf(alpha / 2, successes, n - successes + 1))
            bounds.append(beta.ppf(1 - alpha / 2, successes + 1, n - successes))

        if not bounds:
            bounds = [0, 0]
        return bounds



import math
from estimators.slates import base

# PseudoInverse estimator for slate recommendation. The following implements the
# case for a Cartesian product when mu is a product distribution. This can be
# seen in example 4 of the paper.
# https://arxiv.org/abs/1605.04812

class Estimator(base.Estimator):
    def __init__(self):
        self.data = {'n':0.,'N':0}

    def add_example(self, p_logs, r, p_preds, count=1):
        """Expects lists for logged probabilities and predicted probabilities. These should correspond to each slot.
        This function is implemented under the simplifying assumptions of
        example 4 in the paper 'Off-policy evaluation for slate recommendation'
        where the slate space is a cartesian product and the logging policy is a
        product distribution"""
        if not isinstance(p_logs, list) or not isinstance(p_preds, list):
            raise('Error: p_logs and p_preds must be lists')

        if(len(p_logs) != len(p_preds)):
            raise('Error: p_logs and p_preds must be the same length, found {} and {} respectively'.format(len(p_logs), len(p_preds)))

        self.data['N'] += count
        p_over_ps = 0
        num_slots = len(p_logs)
        for p_log, p_pred in zip(p_logs, p_preds):
            p_over_ps += p_pred/p_log
        p_over_ps -= num_slots - 1

        if r != 0:
            self.data['n'] += r*p_over_ps*count

    def get(self):
        if self.data['N'] == 0:
            raise('Error: No data point added')

        return self.data['n']/self.data['N']



from estimators.bandits import base

class Estimator(base.Estimator):

    def __init__(self):
        ################################# Aggregates quantities #########################################
        #
        # 'n':   IPS of numerator
        # 'N':   total number of samples in bin from log (IPS = n/N)
        # 'd':   IPS of denominator (SNIPS = n/d)
        #
        #################################################################################################

        self.data = {'n':0.,'N':0,'d':0.}

    def add_example(self, p_log, r, p_pred, count=1):
        self.data['N'] += count
        if p_pred > 0:
            p_over_p = p_pred/p_log
            self.data['d'] += p_over_p*count
            if r != 0:
                self.data['n'] += r*p_over_p*count

    def get(self):
        if self.data['N'] == 0:
            raise('Error: No data point added')

        if self.data['d'] != 0:
            return self.data['n']/self.data['d']
        else:
            return 0



import math
from estimators.bandits import base

class Interval(base.Interval):

    def __init__(self):
        ################################# Aggregates quantities #########################################
        #
        # 'n':   IPS of numerator
        # 'N':   total number of samples in bin from log (IPS = n/N)
        # 'SoS': sum of squares of numerator's items (needed for Gaussian confidence intervals)
        #
        #################################################################################################

        self.data = {'n':0.,'N':0,'SoS':0}

    def add_example(self, p_log, r, p_pred, count=1):
        self.data['N'] += count
        if p_pred > 0:
            p_over_p = p_pred/p_log
            if r != 0:
                self.data['n'] += r*p_over_p*count
                self.data['SoS'] += ((r*p_over_p)**2)*count

    def get(self, alpha=0.05):
        bounds = []
        num = self.data['n']
        den = self.data['N']
        SoS = self.data['SoS']

        if SoS > 0.0 and den > 1:
            zGaussianCdf = {
              0.25: 1.15,
              0.1: 1.645,
              0.05: 1.96
            }

            variance = (SoS - num * num / den) / (den - 1)
            gaussDelta = zGaussianCdf[alpha] * math.sqrt(variance/den)
            bounds.append(num / den - gaussDelta)
            bounds.append(num / den + gaussDelta)

        if not bounds:
            bounds = [0, 0]
        return bounds



""" Interface for implementation of slates estimator """

from abc import ABC, abstractmethod
from typing import List

class Estimator(ABC):
	""" Interface for implementation of slates estimator """

	@abstractmethod
	def add_example(self, p_logs: List, r: float, p_preds: List, count: float) -> None:
		""" 
		Args:
			p_logs: List of probabilities of the logging policy
			r: reward for choosing an action in the given context
			p_preds: List of predicted probabilities of making decision
			count: weight
		"""
		...

	@abstractmethod
	def get(self) -> float:
		""" Calculates the selected estimator
		Returns:
			The estimator value
		"""
		...

class Interval(ABC):
	""" Interface for implementation of slates estimator interval """

	@abstractmethod
	def add_example(self, p_logs: List, r: float, p_preds: List, count: float) -> None:
		""" 
		Args:
			p_logs: List of probabilities of the logging policy
			r: reward for choosing an action in the given context
			p_preds: List of predicted probabilities of making decision
			count: weight
		"""
		...

	@abstractmethod
	def get(self, alpha: float) -> List:
		""" Calculates the CI
		Args:
			alpha: alpha value
		Returns:
			Returns the confidence interval as list[float]
		"""
		...
